
# Metarace : Cycle Race Abstractions
# Copyright (C) 2012  Nathan Fraser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Printing/Reporting tools.

Create report and write data equivalents to:

	pdf : write to file via Cairo pdf sruface
	gtk.Print : write to printer via Cairo and gtk.PrintOperation
	xls : write xls spreadsheet to file
	txt : write markdown/html to file for web export

The top level report object collects together a pdf style template,
report meta-data and a list of printing sections. Usage:

 - customise template.ini if applicable and include logo svg images in
   event configuration path

 - create printrep object with template

   rep = printrep(template)

 - set metadata on report as required	(may also be set in template)

   rep.strings['title'] = 'title'
   rep.strings['subtitle'] = ...

 - create and add sections to report

   rep.add_section(sec)

 - write PDF to file...

        ofile = os.path.join(configpath, 'output.pdf')
        with open(ofile, 'wb') as f:
            rep.output_pdf(f)

 - for gtk.print:

    def begin_print(self, operation, context, rep):
        rep.start_gtkprint(context.get_cairo_context())
        operation.set_n_pages(rep.get_pages())
        operation.set_unit('points')	# IMPORTANT!

    def draw_print_page(self, operation, context, page_nr, rep):
        rep.set_context(context.get_cairo_context())
        rep.draw_page(page_nr)

 - write out spreadsheet:
 
        ofile = os.path.join(configpath, 'output.xls')
        with open(ofile, 'wb') as f:
            rep.output_xls(f)

 - write out text/html:

        ofile = os.path.join(configpath, 'output.txt')
        with open(ofile, 'wb') as f:
            rep.output_text(f, linkbase, linktypes)

   NOTE: text/html output allows for links to other file types, see
	 testprint for examples.

TODO:

	- list bullet processing to match markdown or item assignment
	- add one test for each section type to testprint app
	- complete xls/text export for all esoteric program elements
	- complete rider info for sprint sections to allow for use in result

	- template editor with preview
	- template 'initialisation' for a project - needs thought
	- modify print elems to create surface patterns, check
	  for storage/embedding

"""

import os
import cairo
import rsvg
import pango
import pangocairo
import datetime
import math
import ConfigParser
import metarace
import csv
import xlwt
from metarace import strops
from metarace import htlib

# xls cell styles
XS_LEFT = xlwt.easyxf()
XS_RIGHT = xlwt.easyxf('align: horiz right')
XS_TITLE = xlwt.easyxf('font: bold on')
XS_SUBTITLE = xlwt.easyxf('font: italic on')
XS_MONOSPACE = xlwt.easyxf('font: name Courier')

# conversions
def pt2mm(pt=1):
    """72pt -> 25.4mm (1 inch)"""
    return float(pt) * 25.4 / 72.0

def mm2pt(mm=1):
    """25.4mm -> 72pt (1 inch)"""
    return float(mm) * 72.0 / 25.4

def in2pt(inval=1):
    """1in -> 72pt"""
    return float(inval) * 72.0

def cm2pt(cm=1):
    """2.54cm => 72pt (1 inch)"""
    return float(cm) * 72.0 / 2.54

def pt2pt(pt=1):
    """Dummy conversion."""
    return pt

BODYFONT = 'TeXGyreTermes 7.4'			# body text
BODYBOLDFONT = 'TeXGyreTermes bold 7.4'		# bold body text
MONOSPACEFONT = 'TexGyreCursor bold 7.4'	# monospaced text
SECTIONFONT = 'TeXGyreAdventor bold 7.4'	# section headings
SUBHEADFONT = 'TeXGyreTermes italic 7.4'	# section subheadings
PROVFONT = 'Nimbus Sans L Bold Condensed 90'	# provisonal underlay
LINE_HEIGHT = mm2pt(5.0)			# body text line height
PAGE_OVERFLOW = mm2pt(2.0)			# section overflow
SECTION_HEIGHT = mm2pt(7.0)			# height of section title


UNITSMAP = { 'mm':mm2pt,
             'cm':cm2pt,
             'in':in2pt,
             'pt':pt2pt, }

def deg2rad(deg=1):
    """convert degrees to radians."""
    return math.pi * float(deg)/180.0

def pi2rad(ang=1):
    """convert multiple of pi to radians."""
    return math.pi * float(ang)

def rad2rad(ang=1):
    """Dummy converter."""
    return ang

ANGUNITSMAP = { 'dg':deg2rad,
                'pi':pi2rad,
                'rd':rad2rad, }

def str2angle(anglestr=None):
    """From degrees, return an angle in radians -2pi -> 2pi"""
    if anglestr is None:
        anglestr = ''
    units = anglestr.strip()[-2:]
    ukey = units.lower()
    val = anglestr
    if ukey not in ANGUNITSMAP:
        ukey = 'dg'
    else:
        val = anglestr.replace(units, '')
    fval = 0.0
    try:
        fval = float(val)
    except ValueError:
        # ignore float value errors
        pass
    return ANGUNITSMAP[ukey](fval)
    
def str2align(alignstr=None):
    """Return an alignment value 0.0 - 1.0."""
    if alignstr is None:
        alignstr = ''
    ret = 0.5
    try:
        ret = float(alignstr)
        if ret < 0.0:
            ret = 0.0
        elif ret > 1.0:
            ret = 1.0
    except ValueError:
        # ignore float value errors
        pass
    return ret

def str2len(lenstr=None):
    """Return a length in points from the supplied string."""
    if lenstr is None:
        lenstr = ''
    units = lenstr.strip()[-2:]
    ukey = units.lower()
    val = lenstr
    if ukey not in UNITSMAP:
        ukey = 'mm'
    else:
        val = lenstr.replace(units, '')
    fval = 0.0
    try:
        fval = float(val)
    except ValueError:
        # ignore float value errors
        pass
    return UNITSMAP[ukey](fval)

def str2dash(dashstr=None):
    ret = None
    if dashstr:
        dvec = dashstr.split()
        rvec = []
        for d in dvec:
            rvec.append(str2len(d))
        if len(rvec) > 0:
            ret = rvec
    return ret

def str2colour(colstr = None):
    """Return a valid colour from supplied string."""
    ret = [0.0, 0.0, 0.0]
    if colstr:
        cvec = colstr.split(',')
        if len(cvec) == 3:
            try:
                for c in range(0,3):
                    ret[c] = float(cvec[c])
                    if ret[c] < 0.0:
                        ret[c] = 0.0
                    elif ret[c] > 1.0:
                        ret[c] = 1.0
            except ValueError:
                pass
    return ret
    
def vecmap(vec=[], maxkey=10):
    """Return a full map for the supplied vector."""
    ret = {}
    for i in range(0,maxkey):
        ret[i] = None
    for i in range(0,len(vec)):
        if vec[i]:
            ret[i] = vec[i].strip()
    return ret

def vecmapstr(vec=[], maxkey=10):
    """Return a full map for the supplied vector, converted to strings."""
    ret = {}
    for i in range(0,maxkey):
        ret[i] = ''
    for i in range(0,len(vec)):
        if vec[i]:
            ret[i] = str(vec[i]).strip()
    return ret

def vec2htmlrow(vec=[]):
    rowmap = vecmapstr(vec, 7)
    cols = []
    cols.append(htlib.td(htlib.escapetext(rowmap[0])))	# Rank (left)
    cols.append(htlib.td(htlib.escapetext(rowmap[1]),
                           {'align':'right'}))	# No (right)
    cols.append(htlib.td(htlib.escapetext(rowmap[2])))	# Name (left)
    cols.append(htlib.td(htlib.escapetext(rowmap[3])))	# Cat/Code (left)
    cols.append(htlib.td(htlib.escapetext(rowmap[4]),
                           {'align':'right'}))	# time/gap (right)
    cols.append(htlib.td(htlib.escapetext(rowmap[5]),
                           {'align':'right'}))	# time/gap (right)
    cols.append(htlib.td(htlib.escapetext(rowmap[6])))	# Units (left)
    return htlib.tr(cols)

def vec2line(vec=[]):
    ret = []
    for i in vec:
        if i is not None:
            ret.append(str(i))
        else:
            ret.append('')
    return ' '.join(ret).strip() + '   \n'

class dual_ittt_startlist(object):
    """Two-up time trial for individual riders."""
    def __init__(self):
        self.heading = None
        self.lines = []
        self.fslbl = 'Front Straight'
        self.bslbl = 'Back Straight'
        self.lenstr = ''
        self.xtra1 = ''
        self.xtra2 = ''
        self.lcount = 0
        self.recordstr = ''
        self.h = None
        
    def set_record(self, recstr):
        """Set the Australian record for this event."""
        if recstr:
            if recstr[0].isdigit():
                self.recordstr = 'Australian Record: ' + recstr
            else:	# ASSUME prompt provided
                self.recordstr = recstr
        else:
            self.recordstr = ''

    def set_single(self):
        """Convenience func to make 'single lane'."""
        self.fslbl = ''
        self.bslbl = ''

    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            self.h = report.line_height * len(self.lines)
            if self.bslbl:
                self.h *= 2
            if self.heading:
                self.h += report.section_height
            if self.fslbl or self.bslbl:
                self.h += report.line_height
            if self.xtra1 or self.xtra2:
                self.h += report.line_height
            if self.recordstr:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        # Fit whole section on page
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)
        else:
            # move section onto next page
            return (pagebreak(), self)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        subheadvec = ['', '', '']
        dosubhead = False
        if self.lenstr:
            subheadvec[0] = self.lenstr
            dosubhead = True
        if self.xtra2:
            subheadvec[1] = self.xtra2
            dosubhead = True
        if self.xtra1:
            subheadvec[2] = self.xtra1
            dosubhead = True
        if dosubhead:
            report.text_cent(report.midpagew, report.h, 
                       '\t'.join(subheadvec).strip(), report.fonts['subhead'])
            report.h += report.line_height
        dosubhead = False
        dual = False
        if self.fslbl:
            report.text_cent(report.midpagew-mm2pt(40), report.h, self.fslbl,
                              report.fonts['subhead'])
            dosubhead = True
        if self.bslbl:
            report.text_left(report.midpagew+mm2pt(40), report.h, self.bslbl,
                              report.fonts['subhead'])
            dosubhead = True
            dual = True
        if dosubhead:
            report.h += report.line_height

        hof = report.h
        lineheight = report.line_height
        if dual:
            lineheight *= 2
        for i in self.lines:
            report.ittt_heat(i, hof, dual)
            hof += lineheight
        
        if self.recordstr:
            report.text_cent(report.midpagew, hof, self.recordstr,
                              report.fonts['subhead'])
            hof += report.line_height
        report.h = hof
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        return None

    def draw_text(self, report, f):
        """Output program element in markdown."""
        return ''

class twocol_startlist(object):
    def __init__(self):
        self.heading = None
        self.subheading = None
        self.timestr = None
        self.lines = []
        self.lcount = 0
        self.h = None
        
    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            self.h = report.line_height * math.ceil(0.5 * len(self.lines))
            if self.heading:
                self.h += report.section_height
            if self.subheading:
                self.h += report.line_height
            if self.timestr:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""
        # program event sections do not break ... 
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)
        else:
            if report.pagefrac() < 0.0001:	# avoid error
                raise RuntimeWarning('Section ' + repr(self.heading)
                          + ' will not fit on a page and will not break.')
            # move entire section onto next page
            return (pagebreak(), self)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        if self.subheading:
            report.text_cent(report.midpagew, report.h, self.subheading,
                              report.fonts['subhead'])
            report.h += report.line_height

        colof = report.body_left
        hof = report.h
        collen = int(math.ceil(0.5 * len(self.lines)))
        if len(self.lines) > 0:
            for i in self.lines[0:collen]:
                if len(i) > 2:
                    report.rms_rider(i, colof, hof)
                hof += report.line_height
            hof = report.h
            colof = report.midpagew+mm2pt(2.0)
            for i in self.lines[collen:]:
                if len(i) > 2:
                    report.rms_rider(i, colof, hof)
                hof += report.line_height
        report.h += collen * report.line_height

        if self.timestr:
            baseline = report.get_baseline(report.h)
            report.text_right(report.body_right - mm2pt(21.0), report.h,
                              self.timestr, report.fonts['subhead'])
            report.drawline(report.body_right - mm2pt(20.0), baseline,
                            report.body_right, baseline)
            report.h += report.line_height

        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        row = report.h
        if self.heading:
            worksheet.write(row, 2, self.heading.strip(), XS_TITLE)
            row += 1
        if self.subheading:
            worksheet.write(row, 2,
                 self.subheading.replace('\t', '  ').strip(), XS_SUBTITLE)
            row += 2
        else:
            row += 1	# min one clear row between
 
        if len(self.lines) > 0:
            rows = []
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(vecmapstr(nv, 7))
            for l in rows:
                # todo: apply styles to whole doc?
                worksheet.write(row, 0, l[0], XS_LEFT)
                worksheet.write(row, 1, l[1], XS_RIGHT)
                worksheet.write(row, 2, l[2], XS_LEFT)
                worksheet.write(row, 3, l[3], XS_LEFT)
                worksheet.write(row, 4, l[4], XS_RIGHT)
                worksheet.write(row, 5, l[5], XS_RIGHT)
                worksheet.write(row, 6, l[6], XS_LEFT)
                row += 1
            row += 1
        report.h = row
        return None

    def draw_text(self, report, f):
        """Write out a section in markdown."""
        if self.heading:
            f.write('### ' + self.heading.strip() + '\n\n')
        if self.subheading:
            f.write('#### ' + self.subheading.strip() + '\n\n')
        if len(self.lines) > 0:
            rows = []
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(nv)
            trows = []
            for l in rows:
                trows.append(vec2htmlrow(l))
            f.write(htlib.table(trows, {'border':'0',
                                 'cellspacing':'2', 'cellpadding':'2'}))
            f.write('\n\n')
        return False

class sprintround(object):
    def __init__(self):
        self.heading = None
        self.lenstr = None
        self.xtra1 = None
        self.xtra2 = None
        self.lines = []		 # maps to 'heats'
        self.lcount = 0
        self.h = None
        
    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            self.h = report.line_height * len(self.lines)
            if self.heading:
                self.h += report.section_height
            if self.xtra1 or self.xtra2 or self.lenstr:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        # program event sections do not break
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)
        else:
            # move entire section onto next page
            return (pagebreak(), self)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        subheadvec = ['', '', '']
        dosubhead = False
        if self.lenstr:
            subheadvec[0] = self.lenstr
            dosubhead = True
        if self.xtra2:
            subheadvec[1] = self.xtra2
            dosubhead = True
        if self.xtra1:
            subheadvec[2] = self.xtra1
            dosubhead = True
        if dosubhead:
            report.text_cent(report.midpagew, report.h, 
                       '\t'.join(subheadvec).strip(), report.fonts['subhead'])
            report.h += report.line_height

        hof = report.h
        if len(self.lines) > 0:
            for i in self.lines:
                heat = ''
                if i[0]:
                    heat = i[0]
                if heat:
                    report.text_left(report.body_left, hof,
                                     heat, report.fonts['subhead']) 
                report.rms_rider(i[1], report.body_left + mm2pt(12), hof)
                report.text_cent(report.midpagew - mm2pt(2), hof, 'v',
                                  report.fonts['subhead']) 
                report.rms_rider(i[2], report.midpagew + mm2pt(2), hof)
                time = ''
                if i[3]:
                    time = i[3]
                if time:
                    report.text_left(report.body_right - mm2pt(10), hof,
                                     time, report.fonts['body']) 
                else:
                    baseline = report.get_baseline(hof)
                    report.drawline(report.body_right - mm2pt(10),
                                    baseline,
                                    report.body_right,
                                    baseline)
                hof += report.line_height
        report.h = hof
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        return None

    def draw_text(self, report, f):
        """Output program element in markdown."""
        return ''

class sprintfinal(object):
    def __init__(self):
        self.heading = None
        self.lenstr = None
        self.xtra1 = None
        self.xtra2 = None
        self.lines = []		 # maps to 'contests'
        self.lcount = 0
        self.h = None
        
    def get_h(self, report):
        """Return total height on page of section."""
        if self.h is None or len(self.lines) != self.lcount:
            self.h = report.line_height * 3.0 * len(self.lines)
            if self.heading:
                self.h += report.section_height
            if self.xtra1 or self.xtra2 or self.lenstr:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""
        # program event sections do not break
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)
        else:
            # move entire section onto next page
            return (pagebreak(), self)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        subheadvec = ['', '', '']
        dosubhead = False
        if self.lenstr:
            subheadvec[0] = self.lenstr
            dosubhead = True
        if self.xtra2:
            subheadvec[1] = self.xtra2
            dosubhead = True
        if self.xtra1:
            subheadvec[2] = self.xtra1
            dosubhead = True
        if dosubhead:
            report.text_cent(report.midpagew, report.h, 
                       '\t'.join(subheadvec).strip(), report.fonts['subhead'])
            report.h += report.line_height

        hof = report.h
        if len(self.lines) > 0:
            for i in self.lines:
                hw = mm2pt(20)
                hl = report.midpagew + hw
                h1t = hl + 0.5*hw
                h2t = h1t + hw
                h12 = hl + hw
                h3t = h2t + hw
                h23 = h12 + hw
                hr = hl + 3.0*hw

                # place heat headings
                report.text_cent(h1t, hof, 'Heat 1', report.fonts['subhead']) 
                report.text_cent(h2t, hof, 'Heat 2', report.fonts['subhead']) 
                report.text_cent(h3t, hof, 'Heat 3', report.fonts['subhead']) 
                hof += report.line_height

                heat = ''
                if i[0]:
                    heat = i[0]
                if heat:
                    report.text_left(report.body_left, hof,
                                     heat, report.fonts['subhead']) 

                ht = hof
                bl = report.get_baseline(hof)
                hb = report.get_baseline(hof + report.line_height)
                # draw heat lines
                report.drawline(hl, bl, hr, bl)
                report.drawline(h12, ht, h12, hb)
                report.drawline(h23, ht, h23, hb)

                report.rms_rider(i[1], report.body_left+hw, hof)
                hof += report.line_height
                report.rms_rider(i[2], report.body_left+hw, hof)
                hof += report.line_height
        report.h = hof
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        return None

    def draw_text(self, report, f):
        """Output program element in markdown."""
        return ''

class rttstartlist(object):
    def __init__(self):
        self.heading = None
        self.colheader = None
        self.logo = None
        self.lines = []
        self.lcount = 0
        self.h = None

    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            cvec = self.lines[0:]
            if self.colheader:	# colheader is written out with body
                cvec.append(self.colheader)
            self.h = report.column_height(cvec)
            if self.heading or self.logo:
                self.h += report.section_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""

        # Special case 1: Entire section will fit on page, even if squeezed
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)

        # Special case 2: Not enough space for minimum content
        chk = rttstartlist()
        chk.heading = self.heading
        chk.logo = self.logo
        chk.colheader = self.colheader
        if len(self.lines) == 3: # special case, keep 'threes' together
            chk.lines = self.lines[0:]
        else:
            chk.lines = self.lines[0:2]
        if chk.get_h(report) > remainder:
            # move entire section onto next page
            return (pagebreak(), self)

        # Standard case - section must be broken
        ret = rttstartlist()
        rem = rttstartlist()
        ret.heading = self.heading
        ret.logo = self.logo
        rem.heading = self.heading
        rem.logo = self.logo
        if rem.heading is not None:
            if rem.heading.rfind('(continued)') < 0:
                rem.heading += ' (continued)'
        ret.colheader = self.colheader
        rem.colheader = self.colheader
        seclines = len(self.lines)
        count = 0
        if seclines > 0:
            ret.lines.append(self.lines[0])
            count = 1 # case: 3 lines broken on first line
        while count < seclines:
            if ret.get_h(report) > remainder:
                # pop last line onto rem and break
                rem.lines.append(ret.lines.pop(-1))
                break
            elif seclines - count <= 2: # push min 2 lines over to next page
                break
            ret.lines.append(self.lines[count])
            count += 1
        while count < seclines:
            rem.lines.append(self.lines[count])
            count += 1
        return(ret, rem)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        if self.logo is not None:
            ### make image elem and then draw it
            pass
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(self.colheader)
            rows.extend(self.lines)
            # just hard-code cols for now, later do a colspec?
            report.output_column(rows, 0, 'r', report.col_oft_name)
            report.output_column(rows, 4, 'l', report.col_oft_name + mm2pt(2))
            report.output_column(rows, 1, 'r', report.col_oft_name + mm2pt(16))
            new_h = report.output_column(rows, 2, 'l', 
                                                report.col_oft_name+ mm2pt(20))
            report.output_column(rows, 3, 'l', report.col_oft_cat+ mm2pt(20))
            report.output_column(rows, 5, 'r', report.col_oft_xtra)
            report.h += new_h
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        row = report.h
        if self.heading is not None:
            worksheet.write(row, 2, self.heading.strip(), XS_TITLE)
            row += 2
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(vecmapstr(self.colheader,7))
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(vecmapstr(nv, 7))
            for l in rows:
                # todo: apply styles to whole doc?
                worksheet.write(row, 0, l[0], XS_LEFT)
                worksheet.write(row, 1, l[1], XS_RIGHT)
                worksheet.write(row, 2, l[2], XS_LEFT)
                worksheet.write(row, 3, l[3], XS_LEFT)
                worksheet.write(row, 4, l[4], XS_RIGHT)
                worksheet.write(row, 5, l[5], XS_RIGHT)
                worksheet.write(row, 6, l[6], XS_LEFT)
                row += 1
            row += 1
        report.h = row
        return None

    def draw_text(self, report, f):
        """Write out a section in markdown."""
        if self.heading is not None:
            f.write('### ' + self.heading.strip() + '\n\n')
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(self.colheader)
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(nv)
            trows = []
            for l in rows:
                trows.append(vec2htmlrow(l))
            f.write(htlib.table(trows, {'border':'0',
                                 'cellspacing':'2', 'cellpadding':'2'}))
            f.write('\n\n')

class bullet_text(object):
    """List of bullet items, each one a non-breaking pango para."""
    def __init__(self):
        self.heading = None	# scalar
        self.subheading = None	# scalar
        self.lines = []		# list of sections: [bullet,para]
        self.lcount = 0		# last count of lines/len
        self.bullet = '*'	# bullet type	?is this the way?
        self.width = None	# allow override of width
        self.h = None		# computed height on page

    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            if self.width is None:	# override by caller allowed
                self.width = report.body_width - mm2pt(15+10)
            self.h = 0
            if self.heading:
                self.h += report.section_height
            if self.subheading:
                self.h += report.line_height
            for line in self.lines:
                bh = report.line_height
                ph = 0
                if line[1]:	# empty or none not drawn
                    ph = report.paragraph_height(line[1], self.width)
                self.h += max(bh, ph)	# enforce a minimum item height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""

        # Special case 1: Entire section will fit on page
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)

        # Special case 2: Not enough space for minimum content
        chk = bullet_text()
        chk.heading = self.heading
        chk.subheading = self.subheading
        chk.lines = self.lines[0:1]	# minimum one item before break
        if chk.get_h(report) > remainder:
            # move entire section onto next page
            return (pagebreak(), self)

        # Standard case - section crosses page break, determines
        # ret: content on current page
        # rem: content on subsequent pages
        ret = bullet_text()
        rem = bullet_text()
        ret.heading = self.heading
        ret.subheading = self.subheading
        rem.heading = self.heading
        rem.subheading = self.subheading
        if rem.heading is not None:
            if rem.heading.rfind('(continued)') < 0:
                rem.heading += ' (continued)'
        ret.bullet = self.bullet
        rem.bullet = self.bullet
        seclines = len(self.lines)
        count = 0
        if seclines > 0:
            ret.lines.append(self.lines[0])
            count = 1 # case: min one line before break
        while count < seclines: 	# visit every item
            if ret.get_h(report) > remainder:
                # if overflow, undo last item and fall out to remainder
                rem.lines.append(ret.lines.pop(-1))
                break
            elif seclines - count <= 1:	
                break	# hanging item check (rm=1)
            ret.lines.append(self.lines[count])
            count += 1
        while count < seclines:
            # collect all remainder items in rem
            rem.lines.append(self.lines[count])
            count += 1
        return(ret, rem)

    def draw_pdf(self, report):
        """Output the bullet list to page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        if self.subheading is not None:
            report.text_cent(report.midpagew, report.h, self.subheading,
                           report.fonts['subhead'])
            report.h += report.line_height
        if len(self.lines) > 0:
            if self.width is None:	# override by caller allowed
                self.width = report.body_width - mm2pt(15+10)
            for l  in self.lines:
                bstr = self.bullet
                if l[0] is not None:
                    bstr = l[0]		# allow override even with ''
                # draw bullet
                bh = report.line_height	# minimum item height is one line
                if bstr:
                    report.text_left(report.body_left+mm2pt(5.0), report.h,
                                     bstr, report.fonts['body'])
                # draw para
                ph = 0
                if l[1]:	# allow empty list item
                    ph = report.text_para(report.body_left+mm2pt(15.0),
                                             report.h, l[1],
                                             report.fonts['body'], self.width)
                report.h += max(ph, bh)
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        return None	# TODO

    def draw_text(self, report, f):
        """Write out a section in markdown."""
        if self.heading is not None:
            f.write('### ' + self.heading.strip() + '\n\n')
        if self.subheading is not None:
            f.write('#### ' + self.subheading.strip() + '\n\n')
        if len(self.lines) > 0:
            for l in self.lines:
                bstr = ''
                if l[0]:
                    bstr = '('+l[0]+') '
                if l[1]: 
                    bstr += l[1]
                f.write('* ' + bstr.rstrip() + '\n')
            f.write('\n\n')

class preformat_text(object):
    """Block of pre-formatted/monospaced plain text."""
    def __init__(self):
        self.heading = None	# scalar
        self.subheading = None	# scalar
        self.colheader = None	# scalar
        self.lines = []		# list of scalars
        self.lcount = 0		# last count of lines/len
        self.h = None		# computed height on page

    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            cvec = self.lines[0:]
            if self.colheader:	# colheader is written out with body
                cvec.append(self.colheader)
            self.h = report.preformat_height(cvec)
            if self.heading:
                self.h += report.section_height
            if self.subheading:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""

        # Special case 1: Entire section will fit on page
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)

        # Special case 2: Not enough space for minimum content
        chk = preformat_text()
        chk.heading = self.heading
        chk.subheading = self.subheading
        chk.colheader = self.colheader
        if len(self.lines) == 3: # special case, keep 'threes' together
            chk.lines = self.lines[0:]
        else:
            chk.lines = self.lines[0:2]
        if chk.get_h(report) > remainder:
            # move entire section onto next page
            return (pagebreak(), self)

        # Standard case - section crosses page break, determines
        # ret: content on current page
        # rem: content on subsequent pages
        ret = preformat_text()
        rem = preformat_text()
        ret.heading = self.heading
        ret.subheading = self.subheading
        rem.heading = self.heading
        rem.subheading = self.subheading
        if rem.heading is not None:
            if rem.heading.rfind('(continued)') < 0:
                rem.heading += ' (continued)'
        ret.colheader = self.colheader
        rem.colheader = self.colheader
        seclines = len(self.lines)
        count = 0
        if seclines > 0:
            ret.lines.append(self.lines[0])
            count = 1 # case: 3 lines broken on first line
        while count < seclines: # case: push min two lines over break
            if ret.get_h(report) > remainder:
                # pop last line onto rem and break
                rem.lines.append(ret.lines.pop(-1))
                break
            elif seclines - count <= 2: # push min 2 lines over to next page
                break
            ret.lines.append(self.lines[count])
            count += 1
        while count < seclines:
            rem.lines.append(self.lines[count])
            count += 1
        return(ret, rem)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading is not None:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        if self.subheading is not None:
            report.text_cent(report.midpagew, report.h, self.subheading,
                           report.fonts['subhead'])
            report.h += report.line_height
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(self.colheader)
            rows.extend(self.lines)
            ust = '\n'.join(rows)
            report.h += report.text_cent(report.midpagew, report.h, ust,
                               report.fonts['monospace'],
                               halign=pango.ALIGN_LEFT)
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        return None	# TODO

    def draw_text(self, report, f):
        """Write out a section in markdown."""
        if self.heading is not None:
            f.write('### ' + self.heading.strip() + '\n\n')
        if self.subheading is not None:
            f.write('#### ' + self.subheading.strip() + '\n\n')
        if len(self.lines) > 0:
            if self.colheader:
                f.write('\t' + self.colheader.rstrip() + '\n')
            for row in self.lines:
                f.write('\t' + row.rstrip() + '\n')
            f.write('\n\n')

class section(object):
    def __init__(self):
        self.heading = None		# scalar
        self.colheader = None		# scalar
        self.subheading = None		# scalar
        self.units = None		# scalar
        self.lines = []			# list of column lists
        self.lcount = 0			
        self.h = None

    def get_h(self, report):
        """Return total height on page of section on report."""
        if self.h is None or len(self.lines) != self.lcount:
            cvec = self.lines[0:]
            if self.colheader:	# colheader is written out with body
                cvec.append(['-','-','-','-','-','-'])
            self.h = report.column_height(cvec)
            if self.heading:
                self.h += report.section_height
            if self.subheading:
                self.h += report.line_height
            self.lcount = len(self.lines)
        return self.h

    def truncate(self, remainder, report):
        """Return a copy of the section up to page break."""

        # Special case 1: Entire section will fit on page
        if self.get_h(report) <= (remainder + report.page_overflow):
            return (self, None)

        # Special case 2: Not enough space for minimum content
        chk = section()
        chk.heading = self.heading
        chk.subheading = self.subheading
        chk.colheader = self.colheader
        chk.units = self.units
        if len(self.lines) == 3: # special case, keep 'threes' together
            chk.lines = self.lines[0:]
        else:
            chk.lines = self.lines[0:2]
        if chk.get_h(report) > remainder:
            # move entire section onto next page
            return (pagebreak(), self)

        # Standard case - section crosses page break, determines
        # ret: content on current page
        # rem: content on subsequent pages
        ret = section()
        rem = section()
        ret.heading = self.heading
        ret.subheading = self.subheading
        rem.heading = self.heading
        rem.subheading = self.subheading
        if rem.heading is not None:
            if rem.heading.rfind('(continued)') < 0:
                rem.heading += ' (continued)'
        ret.colheader = self.colheader
        rem.colheader = self.colheader
        ret.units = self.units
        rem.units = self.units
        seclines = len(self.lines)
        count = 0
        if seclines > 0:
            ret.lines.append(self.lines[0])
            count = 1 # case: 3 lines broken on first line
        while count < seclines: # case: push min two lines over break
            if ret.get_h(report) > remainder:
                # pop last line onto rem and break
                rem.lines.append(ret.lines.pop(-1))
                break
            elif seclines - count <= 2: # push min 2 lines over to next page
                break
            ret.lines.append(self.lines[count])
            count += 1
        while count < seclines:
            rem.lines.append(self.lines[count])
            count += 1
        return(ret, rem)

    def draw_pdf(self, report):
        """Output a single section to the page."""
        report.c.save()
        if self.heading:
            report.text_cent(report.midpagew, report.h, self.heading,
                           report.fonts['section'])
            report.h += report.section_height
        if self.subheading:
            report.text_cent(report.midpagew, report.h, self.subheading,
                           report.fonts['subhead'])
            report.h += report.line_height
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(self.colheader)
            rows.extend(self.lines)
            # just hard-code cols for now, later do a colspec?
            if self.units:
                ust = self.units
                if self.colheader:
                    ust = '\n'+ust 
                report.text_left(report.col_oft_units, report.h, ust,
                               report.fonts['body'])
            report.output_column(rows, 0, 'l', report.col_oft_rank)
            report.output_column(rows, 1, 'r', report.col_oft_no)
            new_h = report.output_column(rows, 2, 'l', report.col_oft_name)
            report.output_column(rows, 3, 'l', report.col_oft_cat)
            report.output_column(rows, 4, 'r', report.col_oft_time)
            report.output_column(rows, 5, 'r', report.col_oft_xtra)
            report.h += new_h
        report.c.restore()

    def draw_xls(self, report, worksheet):
        """Output program element to excel worksheet."""
        row = report.h
        if self.heading:
            worksheet.write(row, 2, self.heading.strip(), XS_TITLE)
            row += 1
        if self.subheading:
            worksheet.write(row, 2, self.subheading.strip(), XS_SUBTITLE)
            row += 2
        else:
            row += 1
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(vecmapstr(self.colheader,7))
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(vecmapstr(nv, 7))
            if self.units:
                if self.colheader:
                    rows[1][6] = self.units
                else:
                    rows[0][6] = self.units
            for l in rows:
                # todo: apply styles to whole doc?
                worksheet.write(row, 0, l[0], XS_LEFT)
                worksheet.write(row, 1, l[1], XS_RIGHT)
                worksheet.write(row, 2, l[2], XS_LEFT)
                worksheet.write(row, 3, l[3], XS_LEFT)
                worksheet.write(row, 4, l[4], XS_RIGHT)
                worksheet.write(row, 5, l[5], XS_RIGHT)
                worksheet.write(row, 6, l[6], XS_LEFT)
                row += 1
            row += 1
        report.h = row
        return None

    def draw_text(self, report, f):
        """Write out a section in markdown."""
        if self.heading:
            f.write('### ' + self.heading.strip() + '\n\n')
        if self.subheading:
            f.write('#### ' + self.subheading.strip() + '\n\n')
        if len(self.lines) > 0:
            rows = []
            if self.colheader:
                rows.append(self.colheader)
            for r in self.lines:
                nv = r[0:6]
                if len(nv) == 2:
                    nv = [nv[0], None, nv[1]]
                rows.append(nv)
            if self.units:
                if self.colheader:
                    rows[1].append(self.units)
                else:
                    rows[0].append(self.units)
            trows = []
            for l in rows:
                trows.append(vec2htmlrow(l))
            f.write(htlib.table(trows, {'border':'0',
                                 'cellspacing':'2', 'cellpadding':'2'}))
            f.write('\n\n')
        return None

class pagebreak(object):
    """Dummy 'section' for page breaks."""
    def __init__(self, threshold=0.3):
        self.threshold = threshold

    def set_threshold(self, threshold=0.3):
        self.threshold = float(threshold)

    def get_threshold(self):
        return self.threshold

class image_elem(object):
    """Place an SVG image on the page."""
    def __init__(self, x1=None, y1=None, x2=None, y2=None,
                        halign=None, valign=None, source=None):
        if halign is None:
            halign = 0.5
        if valign is None:
            valign = 0.5
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2
        self.halign = halign
        self.valign = valign
        self.xof = 0.0
        self.yof = 0.0
        self.sf = 1.0
        self.set_source(source)

    def set_source(self, source=None):
        self.source = source
        if self.source is not None:
            # Pre-compute geometry
            bw = self.x2 - self.x1
            bh = self.y2 - self.y1
            if math.fabs(bh) < 0.01:	# avoid div zero
                bh += 1.0	# but normally an error?
            ab = bw / bh
            iw = float(self.source.props.width)
            ih = float(self.source.props.height)
            ai = iw / ih
            xoft = 0.0
            yoft = 0.0
            sf = 1.0
            if ai > ab:     # 'wider' than box, scale to box w
                # xoft will be 0 for all aligns
                sf = bw / iw
                yoft = self.valign * (bh - ih * sf)
            else:           # 'higher' than box, scale to box h
                # yoft will be 0 for all aligns
                sf = bh / ih
                xoft = self.halign * (bw - iw * sf)
            self.sf = sf
            self.xof = self.x1 + xoft
            self.yof = self.y1 + yoft

    def draw(self, c, p):
        if self.source is not None:
            c.save()
            c.translate(self.xof, self.yof)
            c.scale(self.sf, self.sf)
            self.source.render_cairo(c)
            c.restore()

class arc_elem(object):
    """Pace an optionally shaded arc on the page."""
    def __init__(self, cx=None, cy=None, r=None, a1=None, a2=None, 
                       fill=None, width=None, colour=None, dash=None):
        self.cx = cx
        self.cy = cy
        self.r = r
        self.a1 = a1
        self.a2 = a2
        self.fill = fill
        self.width = width
        self.colour = colour
        self.dash = dash
    def draw(self, c, p):
        c.save()
        #c.move_to(self.cx, self.cy)
        c.new_sub_path()
        c.arc(self.cx, self.cy, self.r, self.a1, self.a2)
        outline = False
        if self.width is not None:
            outline = True
            c.set_line_width(self.width)
        if self.fill is not None:
            c.set_source_rgb(self.fill[0], self.fill[1], self.fill[2])
            if outline:
                c.fill_preserve()
            else:
                c.fill()
        if outline:
            if self.colour is not None:
                c.set_source_rgb(self.colour[0], self.colour[1], self.colour[2])
            if self.dash is not None:
                c.set_dash(self.dash)
            c.stroke()
        c.restore()

class box_elem(object):
    """Place an optionally shaded box on the page."""
    def __init__(self, x1=None, y1=None, x2=None, y2=None, fill=None,
                        width=None, colour=None, dash=None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.fill = fill
        self.width = width
        self.colour = colour
        self.dash = dash

    def draw(self, c, p):
        c.save()
        c.move_to(self.x1, self.y1)
        c.line_to(self.x2, self.y1)
        c.line_to(self.x2, self.y2)
        c.line_to(self.x1, self.y2)
        c.close_path()
        outline = False
        if self.width is not None:
            outline = True
            c.set_line_width(self.width)
        if self.fill is not None:
            c.set_source_rgb(self.fill[0], self.fill[1], self.fill[2])
            if outline:
                c.fill_preserve()
            else:
                c.fill()
        if outline:
            if self.colour is not None:
                c.set_source_rgb(self.colour[0], self.colour[1], self.colour[2])
            if self.dash is not None:
                c.set_dash(self.dash)
            c.stroke()
        c.restore()

class line_elem(object):
    """Places a line on the page."""
    def __init__(self, x1=None, y1=None, x2=None, y2=None,
                        width=None, colour=None, dash=None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.width = width
        self.colour = colour
        self.dash = dash

    def draw(self, c, p):
        c.save()
        if self.width is not None:
            c.set_line_width(self.width)
        if self.colour is not None:
            c.set_source_rgb(self.colour[0], self.colour[1], self.colour[2])
        if self.dash is not None:
            c.set_dash(self.dash)
        c.move_to(self.x1, self.y1)
        c.line_to(self.x2, self.y2)
        c.stroke()
        c.restore()

class text_elem(object):
    """Places string of text on the page."""
    def __init__(self, x=None, y=None, align=None, font=None,
                        colour=None, source=None, report=None):
        self.x = x
        self.y = y
        self.align = align
        self.font = font
        self.colour = colour
        self.source = source
        self.report = report
    def draw(self, c, p):
        msg = None
        if self.source:
            if self.source in self.report.strings:
                if self.report.strings[self.source]:
                    msg = self.report.strings[self.source]
            else:
                msg = self.source
        if msg:
            c.save()
            l = p.create_layout()
            if self.font is not None:
                l.set_font_description(self.font)
            if self.colour is not None:
                c.set_source_rgb(self.colour[0], self.colour[1], self.colour[2])
            l.set_text(msg)
            (tw,th) = l.get_pixel_size()
            c.move_to(self.x-(self.align * tw), self.y)
            p.update_layout(l)
            p.show_layout(l)
            c.restore()

class group_elem(object):
    """Place each defined element on the page."""
    def __init__(self, report=None, elems=[]):
        self.report = report
        self.elems = elems
        self.indraw = False
    def draw(self, c, p):
        if self.indraw:
            return	# Ignore recursion
        self.indraw = True
        c.save()
        for e in self.elems:
            if e in self.report.elements:
                self.report.elements[e].draw(c,p)
        c.restore()
        self.indraw = False

class printrep(object):
    """PDF/GTKPrint Report class."""
    def __init__(self, template=None, path=None):

        # Default asset path
        if path is not None:
            self.path = path
        else:
            self.path = '.'

        # load template	-> also declares page geometry variables
        self.loadconfig(template)

        # Status and context values
        self.provisional = False
        self.pagemarks = False
        self.s = None
        self.c = None
        self.p = None	# these are filled as required by the caller
        self.h = None		# position on page during write
        self.curpage = None	# current page in report
        self.sections = []	# source section data
        self.pages = []		# paginated sections

        # temporary col offset values...
        self.col_oft_rank = self.body_left		# left align
        self.col_oft_no = self.body_left + mm2pt(16)	# right align
        self.col_oft_name = self.body_left + mm2pt(18)	# left align
        self.col_oft_cat = self.body_right - mm2pt(56)	# ~left align
        self.col_oft_time = self.body_right - mm2pt(20)	# right align
        self.col_oft_xtra = self.body_right - mm2pt(2) # right align
        self.col_oft_units = self.body_right - mm2pt(1)	# left

    def reset_geometry(self, width=None, height=None,
                             sidemargin=None, endmargin=None,
                             printmargin=None):
        """Overwrite any new values and then compute page geometry."""
        if width is not None:
            self.pagew = width
        if height is not None:
            self.pageh = height
        if sidemargin is not None:
            self.sidemargin = sidemargin
        if endmargin is not None:
            self.endmargin = endmargin
        if printmargin is not None:
            self.printmargin = printmargin

        # compute midpage values
        self.midpagew = self.pagew / 2.0
        self.midpageh = self.pageh / 2.0

        # compute body region
        self.printh = self.pageh - self.printmargin - self.printmargin
        self.printw = self.pagew - self.printmargin - self.printmargin
        self.body_left = self.sidemargin
        self.body_right = self.pagew - self.sidemargin
        self.body_width = self.body_right - self.body_left
        self.body_top = self.endmargin
        self.body_bot = self.pageh - self.endmargin
        self.body_len = self.body_bot - self.body_top

    def loadconfig(self, filename=None):
        """Initialise the report template."""

        # Default page geometry
        self.pagew = 595.0
        self.pageh = 842.0
        self.sidemargin = mm2pt(25.5)
        self.endmargin = mm2pt(36.2)
        self.printmargin = mm2pt(5.0)
        self.minbreak = 0.3	# 30% minimum page break threshold

        # Default empty template elements
        self.strings = {}
        self.colours = {}
        self.colours['white'] = [1.0, 1.0, 1.0]
        self.colours['shade'] = [0.9, 0.9, 0.9]
        self.colours['black'] = [0.0, 0.0, 0.0]
        self.elements = {}
        self.fonts = {}
        self.fonts['body'] = pango.FontDescription(BODYFONT)
        self.fonts['bodybold'] = pango.FontDescription(BODYBOLDFONT)
        self.fonts['section'] = pango.FontDescription(SECTIONFONT)
        self.fonts['subhead'] = pango.FontDescription(SUBHEADFONT)
        self.fonts['monospace'] = pango.FontDescription(MONOSPACEFONT)
        self.fonts['provisional'] = pango.FontDescription(PROVFONT)
        self.strings = {}
        self.images = {}
        self.header = []

        # read in from template
        tfile = os.path.join(metarace.UI_PATH, 'template.ini')
        if filename is not None:
            tfile = filename
        cr = ConfigParser.ConfigParser()
        cr.add_section('page')
        cr.add_section('elements')
        cr.add_section('fonts')
        cr.add_section('strings')
        cr.add_section('colours')
        try:
            a = len(cr.read(tfile))
            if a == 0:
                pass	# this is a bad read
        except:
            pass	# this is a library read error problem

        # read in page options
        if cr.has_option('page', 'width'):
            self.pagew = str2len(cr.get('page', 'width'))
        if cr.has_option('page', 'height'):
            self.pageh = str2len(cr.get('page', 'height'))
        if cr.has_option('page', 'sidemargin'):
            self.sidemargin = str2len(cr.get('page', 'sidemargin'))
        if cr.has_option('page', 'endmargin'):
            self.endmargin = str2len(cr.get('page', 'endmargin'))
        if cr.has_option('page', 'printmargin'):
            self.printmargin = str2len(cr.get('page', 'printmargin'))
        if cr.has_option('page', 'minbreak'):
            self.minbreak = str2align(cr.get('page', 'minbreak'))
        self.section_height = SECTION_HEIGHT
        if cr.has_option('page', 'secheight'):
            self.section_height = str2len(cr.get('page', 'secheight'))
        self.line_height = LINE_HEIGHT
        if cr.has_option('page', 'lineheight'):
            self.line_height = str2len(cr.get('page', 'lineheight'))
        self.page_overflow = PAGE_OVERFLOW
        if cr.has_option('page', 'pageoverflow'):
            self.page_overflow = str2len(cr.get('page', 'pageoverflow'))
        self.reset_geometry()
        if cr.has_option('page', 'elements'):
            self.header = cr.get('page', 'elements').split()

        # read in font declarations
        for s in cr.options('fonts'):
            self.fonts[s] = pango.FontDescription(cr.get('fonts', s))
        # read in string declarations
        for s in cr.options('strings'):
            self.strings[s] = cr.get('strings', s)
        # read in colours
        for s in cr.options('colours'):
            self.colours[s] = str2colour(cr.get('colours', s))
        # read in page elements
        for s in cr.options('elements'):
            elem = self.build_element(s, cr.get('elements', s))
            if elem is not None:
                self.elements[s] = elem

    def get_image(self, key=None):
        """Return an image handle or None."""
        ret = None
        if key is not None:
            if key not in self.images:
                fname = os.path.join(self.path, key + '.svg')
                fh = None
                if os.path.isfile(fname):
                    fh = rsvg.Handle(fname)
                self.images[key] = fh
            ret = self.images[key]
        return ret

    def pagepoint(self, pstr, orient='x'):
        """Convert a positional string into an absolute page reference."""
        ret = 0.0
        ref = self.pagew
        mid = self.midpagew
        if orient == 'y':	# vertical orientation
            ref = self.pageh
            mid = self.midpageh

        # special cases - 'mid' and 'max'
        if pstr == 'mid':
            ret = mid
        elif pstr == 'max':
            ret = ref
        else:
            relpos = str2len(pstr)
            if relpos < 0.0:
                ret = ref + relpos	# relative to bottom/right
            else:
                ret = relpos		# relative to top/left
        return ret

    def add_element(self, ekey, estr):
        """Build the element and add it to the page."""
        if ekey not in self.header:
            self.header.append(ekey)
        if ekey in self.elements:
            del self.elements[ekey]
        elem = self.build_element(ekey, estr)
        if elem is not None:
            self.elements[ekey] = elem
        
    def build_element(self, ekey, estr):
        """Build the element and add it to the element map."""
        ret = None
        emap = vecmap(estr.split(','))
  
        etype = emap[0].lower()
        if etype == 'line':
            width = str2len(emap[5])
            colour = None
            if emap[6] and emap[6] in self.colours:
                colour = self.colours[emap[6]]
            dash = str2dash(emap[7])
            x1 = self.pagepoint(emap[1], 'x')
            y1 = self.pagepoint(emap[2], 'y')
            x2 = self.pagepoint(emap[3], 'x')
            y2 = self.pagepoint(emap[4], 'y')
            ret = line_elem(x1, y1, x2, y2,
                                            width, colour, dash)
        elif etype == 'image':
            x1 = self.pagepoint(emap[1], 'x')
            y1 = self.pagepoint(emap[2], 'y')
            x2 = self.pagepoint(emap[3], 'x')
            y2 = self.pagepoint(emap[4], 'y')
            halign = str2align(emap[5])
            valign = str2align(emap[6])
            source = self.get_image(emap[7])
            ret = image_elem(x1, y1, x2, y2,
                               halign, valign, source)
        elif etype == 'box':
            fill = None
            if emap[5] and emap[5] in self.colours:
                fill = self.colours[emap[5]]
            width = str2len(emap[6])
            colour = None
            if emap[7] and emap[7] in self.colours:
                colour = self.colours[emap[7]]
            dash = str2dash(emap[8])
            x1 = self.pagepoint(emap[1], 'x')
            y1 = self.pagepoint(emap[2], 'y')
            x2 = self.pagepoint(emap[3], 'x')
            y2 = self.pagepoint(emap[4], 'y')
            ret = box_elem(x1, y1, x2, y2, fill,
                               width, colour, dash)
        elif etype == 'arc':
            fill = None
            if emap[6] and emap[6] in self.colours:
                fill = self.colours[emap[6]]
            width = str2len(emap[7])
            colour = None
            if emap[8] and emap[8] in self.colours:
                colour = self.colours[emap[8]]
            dash = str2dash(emap[9])
            cx = self.pagepoint(emap[1], 'x')
            cy = self.pagepoint(emap[2], 'y')
            r = str2len(emap[3])
            a1 = str2angle(emap[4])
            a2 = str2angle(emap[5])
            ret = arc_elem(cx, cy, r, a1, a2, fill,
                               width, colour, dash)
        elif etype == 'text':
            x = self.pagepoint(emap[1], 'x')
            y = self.pagepoint(emap[2], 'y')
            align = str2align(emap[3])
            font = None
            if emap[4] and emap[4] in self.fonts:
                font = self.fonts[emap[4]]
            colour = None
            if emap[5] and emap[5] in self.colours:
                colour = self.colours[emap[5]]
            source = None
            if emap[6]:
                source = emap[6].strip()
            ret = text_elem(x, y, align, font, colour,
                                  source, self)
        elif etype == 'group':	# slightly special case
            elist = estr.split(',')[1:]
            glist = []
            for e in elist:
                e = e.strip()
                if e:
                    glist.append(e)	# preserve ordering
            ret = group_elem(self, glist)
        return ret

    def set_path(self, path=None):
        """Set the path for template assets."""
        if path is not None:
            self.path = path
        else:
            self.path = '.'

    def get_pages(self):
        return len(self.pages)

    def add_section(self, sec):
        self.sections.append(sec)

    def set_provisional(self, flag=True):
        self.provisional = flag

    def set_pagemarks(self, flag=True):
        self.pagemarks = flag

    def output_xls(self, file=None):
        """Output xls spreadsheet."""
        wb = xlwt.Workbook()
        sheetname = 'report'
        # Docstring?
        ws = wb.add_sheet(sheetname)
        # column widths
        ws.col(0).width = int(5*256)
        ws.col(1).width = int(5*256)
        ws.col(2).width = int(34*256)
        ws.col(3).width = int(11*256)
        ws.col(4).width = int(9*256)
        ws.col(5).width = int(7*256)
        ws.col(6).width = int(3*256)
        
        title = ''
        for s in ['title', 'subtitle']:
            if s in self.strings and self.strings[s]:
                title += self.strings[s] + ' '
        ws.write(0,2,title,XS_TITLE)
        self.h = 2	# Start of 'document'
        for s in ['datestr', 'docstr', 'diststr', 'commstr', 'orgstr']:
            if s in self.strings and self.strings[s]:
                ws.write(self.h, 2, self.strings[s].strip(), XS_LEFT)
                self.h += 1
        self.h += 1
        if self.provisional:
            ws.write(self.h, 2, 'PROVISIONAL',
                     XS_TITLE)
            self.h += 2

        # output all the sections...
        for s in self.sections:
            if type(s) is not pagebreak:
                s.draw_xls(self, ws)	# call into section to draw
        
        wb.save(file)

    def output_text(self, file=None, linkbase='', linktypes=[]):
        """Output a markdown/text version of the report."""
        cw = file
        # plain text header
        title = ''
        for s in ['title', 'subtitle']:
            if s in self.strings and self.strings[s]:
                title += self.strings[s] + ' '
        cw.write(title.strip())
        cw.write('\n\n')
        for s in ['datestr', 'docstr', 'diststr', 'commstr', 'orgstr']:
            if s in self.strings and self.strings[s]:
                cw.write('* ' + self.strings[s].strip() + '\n')

        if len(linktypes) > 0:
            linkmsg = '* Download as:'
            for xtn in linktypes:
                linkmsg += ' [' + xtn + '](' + linkbase + '.' + xtn + ')'
            cw.write(linkmsg + '\n')
        cw.write('\n')

        if self.provisional:
            cw.write('*PROVISIONAL*\n\n')

        # output a jump trigger...
        cw.write('<!-- Jumper:jump -->\n\n')

        # output all the sections...
        for s in self.sections:
            if type(s) is not pagebreak:
                s.draw_text(self, cw)	# call into section to draw

        cw.write('\n')

    def set_context(self, context):
        self.s = None
        self.c = context
        self.p = pangocairo.CairoContext(self.c)

    def start_gtkprint(self, context):
        """Prepare document for a gtkPrint output."""
        self.s = None
        self.c = context
        self.p = pangocairo.CairoContext(self.c)

        # break report into pages as required
        self.paginate()

        # special case - dangerous
        if len(self.pages) > 0 and len(self.pages[-1]) == 0:
            del self.pages[-1]

    def output_pdf(self, file=None):
        """Prepare document and then output to a PDF surface."""

        # create cairo surface and save contexts
        self.s = cairo.PDFSurface(file, self.pagew, self.pageh)
        self.c = cairo.Context(self.s)
        self.p = pangocairo.CairoContext(self.c)

        # break report into pages as required
        self.paginate()
        # Special case: Do not allow trailing breaks to leave empty page
        #               on end of report
        if len(self.pages) > 0 and len(self.pages[-1]) == 0:
            del self.pages[-1]
        npages = self.get_pages()
        
        # output each page
        for i in range(0, npages):
            self.draw_page(i)
            if i < npages - 1:
                self.c.show_page()	# start a new blank page

        # finalise surface - may be a blank pdf if no content
        self.s.flush()
        self.s.finish()

    def draw_element(self, elem):
        """Draw the named element if it is defined."""
        if elem in self.elements:
            self.elements[elem].draw(self.c, self.p)
        else:
            pass

    def draw_page(self, page_nr):
        """Draw the current page onto current context."""

        # clip page print extents
        self.c.save()
        self.c.rectangle(self.printmargin, self.printmargin, 
                         self.printw, self.printh)
        self.c.clip()

        # initialise status values
        self.curpage = page_nr + 1
        self.h = self.body_top
        self.strings['pagestr'] = 'Page ' + str(self.curpage)
        if self.get_pages() > 0:
            self.strings['pagestr'] += ' of ' + str(self.get_pages())

        # draw page template
        if self.provisional:
            self.draw_provisional()
        for e in self.header:
            self.draw_element(e)

        # draw page content
        if self.get_pages() > page_nr:
            for s in self.pages[page_nr]:
                s.draw_pdf(self)	# call into section to draw
                self.h += self.line_height # inter-section gap
                #self.draw_section(s)

        # if requested, overlay page marks
        if self.pagemarks:
            self.draw_pagemarks()

        # restore context
        self.c.restore()

    def paragraph_height(self, text, width=None):
        """Determine height of a paragraph at the desired width."""
        ret = 0
        if width is None:
            width = self.body_width
        l = self.p.create_layout()
        if self.fonts['body'] is not None:
            l.set_font_description(self.fonts['body'])
        l.set_width(int(pango.SCALE * width + 1))
        l.set_wrap(pango.WRAP_WORD_CHAR)
        l.set_alignment(pango.ALIGN_LEFT)
        l.set_text(text)
        (tw,th) = l.get_pixel_size()
        ret = th
        return ret
    
    def preformat_height(self, rows):
        """Determine height of a block of preformatted text."""
        ret = 0
        if len(rows) > 0:
            ostr = 'M' + 'L\n'*(len(rows)-1) + 'LM'
            l = self.p.create_layout()
            if self.fonts['monospace'] is not None:
                l.set_font_description(self.fonts['monospace'])
            l.set_text(ostr)
            (tw,th) = l.get_pixel_size()
            ret = th
        return ret

    def column_height(self, rows):
        """Determine height of column."""
        ret = 0
        rvec = []
        for r in rows:
            nval = 'M'
            rvec.append(nval)
        if len(rvec) > 0:
            l = self.p.create_layout()
            if self.fonts['body'] is not None:
                l.set_font_description(self.fonts['body'])
            l.set_text('\n'.join(rvec))
            (tw,th) = l.get_pixel_size()
            ret = th
        return ret

    def output_column(self, rows, col, align, oft):
        """Draw a single column."""
        ret = 0
        rvec = []
        oneval = False
        for r in rows:
            nval = ''
            if len(r) == 2:
                # special case... 
                if col == 2 and r[1]:
                    nval = str(r[1])
                    oneval = True
                elif col == 1 and r[0]:
                    nval = str(r[0])
                    oneval = True
            elif len(r) > col and r[col]:
                nval = str(r[col])
                oneval = True
            rvec.append(nval)
        if oneval:
            if align == 'l':
                ret = self.text_left(oft, self.h, '\n'.join(rvec),
                                     self.fonts['body'])
            else:
                ret = self.text_right(oft, self.h, '\n'.join(rvec),
                                     self.fonts['body'])
        return ret
            
    def newpage(self):
        """Called within paginate to add new page."""
        #self.h = 0.0
        self.h = self.body_top
        curpage = []
        self.pages.append(curpage)
        return curpage

    def pagerem(self):
        """Within paginate, remaining vertical space on current page."""
        return self.body_bot - self.h
        #return self.body_len - self.h

    def pagefrac(self):
        """Within paginate, fractional position on page."""
        return (self.h-self.body_top) / self.body_len

    def paginate(self):
        """Scan report content and paginate sections."""

        # initialise
        self.pages = []
        curpage = self.newpage()

        for r in self.sections:
            s = r
            while s is not None:
                if type(s) is pagebreak:
                    bpoint = s.get_threshold()
                    if bpoint is None:
                        bpoint = self.minbreak
                    if self.pagefrac() > bpoint:
                        curpage = self.newpage() # conditional break
                    s = None
                else:
                    (o, s) = s.truncate(self.pagerem(), self)
                    if type(o) is pagebreak:
                        curpage = self.newpage() # mandatory break
                    else:
                        curpage.append(o)
                        self.h += o.get_h(self)
                        if s is not None:	# section broken to new page
                            curpage = self.newpage()
                        else:
                            self.h += self.line_height # inter sec gap
        
    def draw_pagemarks(self):
        """Draw page layout markings on current page."""
        dash = [mm2pt(1)]
        self.c.save()	# start group
        self.c.set_dash(dash)
        self.c.set_line_width(0.5)
        self.c.set_source_rgb(0.0, 0.0, 1.0)

        # Lay lines
        self.c.move_to(0, 0)
        self.c.line_to(self.pagew, self.pageh)
        self.c.move_to(0, self.pageh)
        self.c.line_to(self.pagew, 0)

        # Page width circles
        self.c.move_to(0, self.midpagew)
        self.c.arc(self.midpagew, self.midpagew, self.midpagew, math.pi, 0.0)
        self.c.move_to(self.pagew, self.pageh-self.midpagew)
        self.c.arc(self.midpagew, self.pageh-self.midpagew, self.midpagew,
                     0.0, math.pi)
        self.c.stroke()

        # Body cropping from page geometry
        self.c.set_source_rgb(0.0, 1.0, 0.0)
        self.c.move_to(0, self.body_top)
        self.c.line_to(self.pagew, self.body_top)
        self.c.move_to(self.body_left,0)
        self.c.line_to(self.body_left,self.pageh)
        self.c.move_to(0, self.body_bot)
        self.c.line_to(self.pagew, self.body_bot)
        self.c.move_to(self.body_right,0)
        self.c.line_to(self.body_right,self.pageh)
        self.c.stroke()

        self.c.restore() # end group

    def get_baseline(self, h):
        """Return the baseline for a given height."""
        return h + 0.8 * self.line_height	# empirical?

    def ittt_lane(self, rvec, w, h):
        """Draw a single lane."""
        baseline = self.get_baseline(h)
        if rvec[0] is None:	# rider no None implies no rider
            self.text_left(w+mm2pt(8), h, '[No Rider]', self.fonts['body'])
        else:
            if rvec[0]:		# non-empty rider no implies full info
                self.text_right(w+mm2pt(7), h, rvec[0], self.fonts['body'])
                self.text_left(w+mm2pt(8), h, rvec[1], self.fonts['body'])
            else:		# otherwise draw placeholder lines
                self.drawline(w, baseline, w+mm2pt(7), baseline)
                self.drawline(w+mm2pt(8), baseline, w+mm2pt(58), baseline)
            self.drawline(w+mm2pt(59), baseline, w+mm2pt(75), baseline)
            
    def ittt_heat(self, hvec, h, dual=False):
        """Output a single time trial heat."""
        ## heat no provided, but ignored for pdf output?
        if dual and hvec[0] and hvec[0] != '0':
            # allow for a heat holder but no text...
            if hvec[0] != '-':
                self.text_left(self.body_left, h, 'Heat ' + hvec[0],
                               self.fonts['subhead'])
            h += self.line_height
        # if riderno None - no rider at straight, else go for placeholder
        if len(hvec) > 3:	# got a front straight
            self.ittt_lane([hvec[1], hvec[2]], self.body_left, h)
        if len(hvec) > 7: 
            if hvec[5] is not None:
                self.text_cent(self.midpagew, h, 'v', self.fonts['subhead'])
            self.ittt_lane([hvec[5], hvec[6]], self.midpagew+mm2pt(5), h)
        elif dual:
            # No Rider ??
            #self.text_cent(self.midpagew, h, 'v', self.fonts['subhead'])
            self.ittt_lane([None, None], self.midpagew+mm2pt(5), h)

    def rms_rider(self, rvec, w, h):
        baseline = self.get_baseline(h)
        if rvec[0] is not None:
            self.text_left(w, h, rvec[0], self.fonts['body'])
        else:
            self.drawline(w, baseline, w+mm2pt(6), baseline)
        doline = True
        if rvec[1]:	# rider no
            self.text_right(w+mm2pt(12.0), h, rvec[1], self.fonts['body'])
            doline = False
        if rvec[2]:	# rider name
            self.text_left(w+mm2pt(13.0), h, rvec[2], self.fonts['body'])
            doline = False
        if doline:
            self.drawline(w+mm2pt(8.0), baseline, w+mm2pt(60), baseline)
        if rvec[3]:	# cat/hcap/draw/etc
            self.text_left(w+mm2pt(70.0), h, rvec[3], self.fonts['subhead'])

    def text_right(self, w, h, msg, font=None):
        l = self.p.create_layout()
        l.set_alignment(pango.ALIGN_RIGHT)
        if font is not None:
            l.set_font_description(font)
        l.set_text(msg)
        (tw,th) = l.get_pixel_size()
        self.c.move_to(w-tw, h)
        self.p.update_layout(l)
        self.p.show_layout(l)
        return th

    def drawline(self, x1, y1, x2, y2, width=0.5):
        self.c.save()
        self.c.set_line_width(width)
        self.c.move_to(x1, y1)
        self.c.line_to(x2, y2)
        self.c.stroke()
        self.c.restore()

    def text_left(self, w, h, msg, font=None, strikethrough=False):
        l = self.p.create_layout()
        l.set_alignment(pango.ALIGN_LEFT)
        if font is not None:
            l.set_font_description(font)
        l.set_text(msg)
        (tw,th) = l.get_pixel_size()
        self.c.move_to(w, h)
        self.p.update_layout(l)
        self.p.show_layout(l)
        if strikethrough:
            self.drawline(w, h+(th/2), w+tw, h+(th/2))
        return th

    def text_para(self, w, h, text, font=None, width=None):
        if width is None:
            width = self.body_width
        l = self.p.create_layout()
        if font is not None:
            l.set_font_description(font)
        l.set_width(int(pango.SCALE * width + 1))
        l.set_wrap(pango.WRAP_WORD_CHAR)
        l.set_alignment(pango.ALIGN_LEFT)
        l.set_text(text)
        (tw,th) = l.get_pixel_size()
        self.c.move_to(w, h)
        self.p.update_layout(l)
        self.p.show_layout(l)
        return th

    def text_cent(self, w, h, msg, font=None, halign=pango.ALIGN_CENTER):
        l = self.p.create_layout()
        l.set_alignment(halign)
        if font is not None:
            l.set_font_description(font)
        l.set_text(msg)
        (tw,th) = l.get_pixel_size()
        self.c.move_to(w-(0.5 * tw), h)
        self.p.update_layout(l)
        self.p.show_layout(l)
        return th

    def text_path(self, w, h, msg, font=None):
        l = self.p.create_layout()
        if font is not None:
            l.set_font_description(font)
        l.set_text(msg)
        (tw,th) = l.get_pixel_size()
        self.c.move_to(w-(0.5 * tw), h)
        self.p.update_layout(l)
        self.p.layout_path(l)
        self.c.fill()
        return th

    def draw_provisional(self):
        self.c.save()
        self.c.set_source_rgb(1.0,1.0,1.0)
        self.text_cent(self.midpagew, self.body_top - mm2pt(5), 
                       'PROVISIONAL', self.fonts['body'])
        self.c.set_source_rgb(0.90, 0.90, 0.90)
        self.c.rectangle(self.body_left-20, self.body_top-20,
                         self.body_right - self.body_left + 40,
                         self.body_bot - self.body_top + 40)
        self.c.clip()
        self.c.translate(self.midpagew, self.midpageh)
        self.c.rotate(0.95532)

        self.text_path(0, -380,
          'PROVISIONAL\nPROVISIONAL\nPROVISIONAL\nPROVISIONAL\nPROVISIONAL',
                       self.fonts['provisional'])
        self.c.restore()

if __name__ == "__main__":
    print('Use testprint included in main metarace distribution for testing.')
